from __future__ import annotations

from os import getenv
from typing import Literal

import anthropic
import openai
from pydantic import BaseModel, PrivateAttr, model_validator

from strot.analyzer.utils import encode_image, extract_json, guess_image_type

LLMProvider = Literal["openai", "anthropic", "groq", "open-router"]


class LLMInput(BaseModel):
    """
    Input to the LLM: text prompt and optional image bytes.

    Raises:
        ValueError: If prompt is empty.
        ValueError: If image type could not be guessed.
    """

    prompt: str
    image: bytes | None = None

    _img_type: str = PrivateAttr(default="")

    class Config:
        extra = "ignore"
        frozen = True

    @model_validator(mode="after")
    def validate(self) -> LLMInput:
        """Validate the input"""

        # validate prompt not empty
        if not self.prompt.strip():
            raise ValueError("Prompt cannot be empty")

        # validate image type if not empty
        if self.image:
            # this raises ValueError if image type could not be guessed
            self._img_type = guess_image_type(self.image)

        return self


class LLMCompletion(BaseModel):
    """
    Completion from an LLM: content string plus token metadata.
    """

    class Config:
        frozen = True

    value: str
    input_tokens: int
    output_tokens: int
    provider: LLMProvider
    model: str


class LLMClient:
    def __init__(
        self,
        *,
        provider: LLMProvider,
        model: str,
        api_key: str | None = None,
        cost_per_1m_input: float,
        cost_per_1m_output: float,
    ):
        self.__provider = provider.lower()
        self.__model = model
        self.__cost_per_1m_input = cost_per_1m_input
        self.__cost_per_1m_output = cost_per_1m_output

        client: anthropic.AsyncClient | openai.AsyncClient
        match self.__provider:
            case "anthropic":
                client = anthropic.AsyncClient(api_key=api_key)
            case "openai" | "groq" | "open-router":
                base_url = None
                if self.__provider == "groq":
                    base_url = "https://api.groq.com/openai/v1/"
                    api_key = api_key or getenv("GROQ_API_KEY")
                elif self.__provider == "open-router":
                    base_url = "https://openrouter.ai/api/v1/"
                    api_key = api_key or getenv("OPENROUTER_API_KEY")

                client = openai.AsyncClient(api_key=api_key, base_url=base_url)
            case _:
                raise ValueError(f"Unsupported provider: {self.__provider}")

        self.__client = client

    @property
    def provider(self) -> LLMProvider:
        return self.__provider

    @property
    def model(self) -> str:
        return self.__model

    def calculate_cost(self, input_tokens: int, output_tokens: int) -> float:
        """
        Compute total cost given rates per million tokens.
        """
        return (
            input_tokens / 1_000_000 * self.__cost_per_1m_input + output_tokens / 1_000_000 * self.__cost_per_1m_output
        )

    async def get_completion(self, input: LLMInput, *, json: bool = False) -> LLMCompletion:
        """
        Get LLM completion.

        Args:
            input: Input to the LLM: text prompt and optional image bytes.
            json: Whether to return JSON completion (default: False)

        Returns:
            LLMCompletion: LLM completion.
        """
        if self.provider == "anthropic":
            completion = await self.__request_anthropic_client(input, json=json)
        else:
            completion = await self.__request_openai_client(input, json=json)

        return completion

    async def __request_openai_client(self, input: LLMInput, json: bool) -> LLMCompletion:
        """
        Request completion from OpenAI client.

        Args:
            input: LLMInput to the LLM: text prompt and optional image bytes.
            json: Whether to return JSON completion

        Returns:
            LLMCompletion: LLM completion.
        """
        if image := input.image:
            messages = [
                {
                    "role": "user",
                    "content": [
                        {"type": "text", "text": input.prompt},
                        {
                            "type": "image_url",
                            "image_url": {"url": f"data:image/{input._img_type};base64,{encode_image(image)}"},
                        },
                    ],
                }
            ]
        else:
            messages = [{"role": "user", "content": input.prompt}]

        response_format = openai._types.NOT_GIVEN
        if json:
            response_format = {"type": "json_object"}

        chat_completion = await self.__client.chat.completions.create(
            model=self.model, messages=messages, response_format=response_format
        )

        value = chat_completion.choices[0].message.content
        return LLMCompletion(
            value=extract_json(value) if json else value,
            input_tokens=chat_completion.usage.prompt_tokens,
            output_tokens=chat_completion.usage.completion_tokens,
            provider=self.provider,
            model=self.model,
        )

    async def __request_anthropic_client(self, input: LLMInput, json: bool) -> LLMCompletion:
        """
        Request completion from Anthropic client.

        Args:
            input: LLMInput to the LLM: text prompt and optional image bytes.
            json: Whether to return JSON completion

        Returns:
            LLMCompletion: LLM completion.
        """
        if image := input.image:
            messages = [
                {
                    "role": "user",
                    "content": [
                        {"type": "text", "text": input.prompt},
                        {
                            "type": "image",
                            "source": {
                                "type": "base64",
                                "media_type": f"image/{input._img_type}",
                                "data": encode_image(image),
                            },
                        },
                    ],
                }
            ]
        else:
            messages = [{"role": "user", "content": input.prompt}]

        if json:
            messages.append({"role": "assistant", "content": "```json"})

        ai_message = await self.__client.beta.messages.create(
            model=self.model, messages=messages, max_tokens=8092, betas=["computer-use-2025-01-24"]
        )

        value = ai_message.content[0].text
        return LLMCompletion(
            value=extract_json(value) if json else value,
            input_tokens=ai_message.usage.input_tokens,
            output_tokens=ai_message.usage.output_tokens,
            provider=self.provider,
            model=self.model,
        )



import logging
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path

from pydantic import BaseModel


class FileHandlerConfig(BaseModel):
    """Configuration for file-based logging handler."""

    directory: str | Path
    """Directory to store the log file."""
    when: str = "midnight"
    """When to rotate the log file."""
    interval: int = 1
    """Interval between rotations."""
    backupCount: int = 0
    """Number of backup files to keep."""
    encoding: str | None = None
    """Encoding to use for the log file."""
    delay: bool = False
    """Delay the creation of the log file."""
    utc: bool = False
    """Whether to use UTC time for the log file."""

    def get_handler(self, logger_name: str) -> logging.Handler:
        """Create and configure a TimedRotatingFileHandler."""
        target_dir = Path(self.directory)
        target_dir.mkdir(parents=True, exist_ok=True)

        file_handler = TimedRotatingFileHandler(
            filename=target_dir / f"{logger_name}.log",
            when=self.when,
            interval=self.interval,
            backupCount=self.backupCount,
            encoding=self.encoding,
            delay=self.delay,
            utc=self.utc,
        )
        file_handler.setFormatter(logging.Formatter("%(message)s"))
        return file_handler



from pathlib import Path
from typing import Any, Literal

from patchright.async_api import Page

from strot.analyzer.schema import Point
from strot.analyzer.utils import text_match_ratio

__all__ = ("Plugin",)


class Plugin:
    """Playwright page plugin for Strot"""

    def __init__(self, page: Page) -> None:
        self._page = page

    async def evaluate(self, expr: str, args=None) -> Any:
        if not await self._page.evaluate(
            "() => window.strotPluginInjected === true",
            isolated_context=False,
        ):
            _ = await self._page.add_script_tag(path=Path(__file__).parent / "inject.js")

        await self._page.wait_for_load_state("domcontentloaded")
        result = await self._page.evaluate(expr, args, isolated_context=False)
        await self._page.wait_for_load_state("domcontentloaded")
        return result

    async def get_selectors_in_view(self) -> set[str]:
        selectors = await self.evaluate(
            """
            () => {
                const elements = window.getElementsInView(window.getElementsInDOM());
                return elements.map(element => window.generateCSSSelector(element));
            }
            """
        )
        return set(selectors)

    async def click_at_point(self, point: Point) -> bool:
        before_selectors = await self.get_selectors_in_view()
        await self._page.mouse.move(point.x, point.y)
        await self._page.mouse.click(point.x, point.y)
        await self._page.wait_for_timeout(2000)
        after_selectors = await self.get_selectors_in_view()

        added = after_selectors - before_selectors
        removed = before_selectors - after_selectors
        return len(added) > 0 or len(removed) > 0

    async def scroll_to_next_view(self, direction: Literal["up", "down"] = "down") -> bool:
        return await self.evaluate("([direction]) => window.scrollToNextView({ direction })", [direction])

    async def get_parent_container(self, text_sections: list[str]) -> str | None:
        containers: list[dict[str, Any]] = await self.evaluate(
            "([sections]) => window.getContainersWithTextSections(sections)", [text_sections]
        )

        if not containers:
            return None

        target_length = len(" ".join(text_sections))
        for container in containers:
            container["match_ratio"] = text_match_ratio(text_sections, container["text"])
            container["text_length"] = len(container["text"])
            container["extra_text_ratio"] = (
                (len(container["text"]) - target_length) / target_length if target_length > 0 else 0
            )

        # Sort by match ratio (descending), then by extra text ratio (ascending - less extra text is better)
        # Then by text length (descending - larger containers preferred if other factors are similar)
        containers.sort(key=lambda x: (-x["match_ratio"], x["extra_text_ratio"], -x["text_length"]))

        if (best_container := containers[0]) and best_container["match_ratio"] > 0.5:
            return best_container["selector"]

    async def get_last_visible_child(self, parent_container_selector: str) -> str | None:
        return await self.evaluate(
            """
            ([selector]) => {
                const parentContainer = document.querySelector(selector);
                if (!parentContainer) {
                    return null;
                }

                return window.getLastVisibleChild(parentContainer);
            }
            """,
            [parent_container_selector],
        )

    async def scroll_to_element(self, selector: str) -> None:
        await self._page.locator(selector).scroll_into_view_if_needed()

    async def take_screenshot(self, type: Literal["png", "jpeg"] = "png") -> bytes:
        return await self._page.screenshot(type=type)



import typing as t
from copy import deepcopy

from jsonref import replace_refs

from strot.type_adapter.utils import (
    drop_titles as drop_titles_fn,
)
from strot.type_adapter.utils import (
    extract_python_object,
    generate_random_suffix,
    transform_to_anthropic_schema,
    transform_to_gemini_schema,
    transform_to_openai_schema,
    update_schema,
)

T_Retval = t.TypeVar("T_Retval")


class TypeAdapter(t.Generic[T_Retval]):
    """
    Schema generation from and data validation against given type.

    A wrapper around Pydantic's `TypeAdapter` to:
    - populate docstring from object type into schema.
    - support schema transformation for different LLM providers.
    """

    @t.overload
    def __init__(self, __type: type[T_Retval]) -> None: ...
    @t.overload
    def __init__(self, __type: t.Callable[..., T_Retval]) -> None: ...

    def __init__(self, __type: type[T_Retval] | t.Callable[..., T_Retval]) -> None:
        from pydantic.fields import FieldInfo
        from pydantic.type_adapter import TypeAdapter as PydanticTypeAdapter

        self.__id = None
        if t.get_origin(__type) is t.Annotated:
            if len(args := t.get_args(__type)) < 2:
                raise TypeError("Annotated types must have more than 1 argument.")
            if isinstance(field := args[1], FieldInfo) and field.alias:
                self.__id = field.alias

        if self.__id is None:
            self.__id = getattr(__type, "__name__", f"id_{generate_random_suffix()}")

        self.ref = __type
        self.__adapter = PydanticTypeAdapter[T_Retval](type=self.ref)
        self.__schema: dict[str, t.Any] | None = None

    @property
    def id(self) -> str:
        return self.__id  # type: ignore[return-value]

    def generate_schema(
        self,
        resolve_refs: bool = False,
        drop_titles: bool = True,
    ) -> dict[str, t.Any]:
        """
        Get the JSON schema for the type.

        The schema is generated on first access and cached for subsequent calls.
        """
        if self.__schema is None:
            base_schema = self.__adapter.json_schema()
            if resolve_refs and "$defs" in base_schema:
                base_schema = replace_refs(base_schema, lazy_load=False)
                _ = base_schema.pop("$defs", None)
            self.__schema = update_schema(self.ref, base_schema)  # type: ignore[arg-type]

        if drop_titles:
            return drop_titles_fn(deepcopy(self.__schema))
        return self.__schema

    @property
    def openai_schema(self) -> dict[str, t.Any]:
        """Get the JsON schema in openai tool calling format."""
        return transform_to_openai_schema(self.id, self.generate_schema(drop_titles=True))

    @property
    def anthropic_schema(self) -> dict[str, t.Any]:
        """Get the JsON schema in anthropic tool calling format."""
        return transform_to_anthropic_schema(self.id, self.generate_schema(drop_titles=True))

    @property
    def gemini_schema(self) -> dict[str, t.Any]:
        """Get the JsON schema in gemini tool calling format."""
        return transform_to_gemini_schema(self.id, self.generate_schema(drop_titles=True))

    def validate_python(self, obj: t.Any, **kwargs: t.Any) -> T_Retval:
        """
        Validate a Python object against the descripted object.

        :param obj: Python object to validate
        :param kwargs: Extra keyword arguments
        :raises ValidationError: If validation fails
        :return: The validated object
        """
        obj = extract_python_object(obj, schema=self.generate_schema(), value_type="python")
        return self.__adapter.validate_python(obj, **kwargs)

    def validate_json(self, data: str | bytes | bytearray, **kwargs: t.Any) -> T_Retval:
        """
        Validate JSON data against the descripted object.

        :param data: JSON data to validate
        :param kwargs: Extra keyword arguments
        :raises ValidationError: If validation fails
        :return: The validated object
        """
        obj = extract_python_object(data, schema=self.generate_schema(), value_type="json")
        return self.__adapter.validate_python(obj, **kwargs)



import io
import re
import threading
import unicodedata
from base64 import b64encode
from typing import Any

import regex
from json_repair import repair_json
from PIL import Image, ImageDraw
from rapidfuzz import fuzz

from strot.analyzer.schema import Pattern, Point, Request

__all__ = (
    "draw_point_on_image",
    "encode_image",
    "extract_json",
    "extract_potential_cursors",
    "get_potential_pagination_parameters",
    "guess_image_type",
    "normalize",
    "text_match_ratio",
    "tokenize",
)


def normalize(text: str) -> str:
    """Unicode normalization and case folding."""
    return unicodedata.normalize("NFKC", text).casefold()


def tokenize(text: str) -> list[str]:
    """
    Tokenize text into words in a language-agnostic way using Unicode-aware regex.
    This works for space-delimited and non-space-delimited languages.
    """
    return regex.findall(r"\p{L}+", text)


def text_match_ratio(subtexts: list[str], text: str, *, cutoff: int = 80) -> float:
    """
    Compute the ratio of substrings found in text (exact or fuzzy), supporting Unicode and non-English text.
    """
    norm_text = normalize(text)
    words = tokenize(norm_text)

    match_count = 0
    lock = threading.Lock()

    def check_one(subtext: str):
        nonlocal match_count
        norm_subtext = normalize(subtext)

        # Exact match (substring)
        if norm_subtext in norm_text:
            found = True
        else:
            # Fuzzy match with words in text
            found = any(fuzz.ratio(norm_subtext, w, score_cutoff=cutoff) for w in words)

        if found:
            with lock:
                match_count += 1

    threads: list[threading.Thread] = []
    for subtext in subtexts:
        t = threading.Thread(target=check_one, args=(subtext,))
        t.start()
        threads.append(t)

    for t in threads:
        t.join()

    return match_count / len(subtexts) if subtexts else 0.0


def extract_json(s: str) -> str:
    """
    Extracts and repairs JSON value from given string.

    Args:
        s: String to extract JSON from
    """
    text = re.sub(r"^```(?:json)?\s*", "", s.strip())
    return repair_json(re.sub(r"\s*```$", "", text))


def guess_image_type(image: bytes) -> str:
    """
    Guess the image type from the image data.

    Args:
        image: Image data to guess the type of

    Raises:
        ValueError: If image type could not be guessed.

    Returns:
        str: Image type (e.g. "png", "jpeg")
    """
    try:
        with Image.open(io.BytesIO(image)) as img:
            img_type = img.format
            if img_type is None:
                raise ValueError("image type could not be guessed")  # noqa: TRY301
            return img_type.lower()
    except Exception as e:
        raise ValueError("image type could not be guessed") from e


def encode_image(image: bytes) -> str:
    """
    Encode image to base64.

    Args:
        image: Image data to encode

    Returns:
        str: Base64 encoded image data
    """
    return b64encode(image).decode("utf-8")


def draw_point_on_image(
    image_bytes: bytes,
    coords: Point,
    radius: int = 5,
    color: "tuple[int, int, int] | str" = "red",
):
    """
    Draw a small circle at the given `(x, y)` coordinates on an image.

    Args:
        image_bytes: Raw image bytes.
        coords: X and Y coordinates (pixels) of the centre of the point.
        radius: Radius of the circle to draw in pixels.
        color: Fill colour for the circle. An `(R, G, B)` tuple or any Pillow-compatible colour string.

    Returns:
        PIL.Image.Image: The image with the point drawn on it.
    """
    img = Image.open(io.BytesIO(image_bytes)).convert("RGBA")

    draw = ImageDraw.Draw(img)

    fill_color = tuple(color) if isinstance(color, tuple) else color

    bbox = (coords.x - radius, coords.y - radius, coords.x + radius, coords.y + radius)
    draw.ellipse(bbox, fill=fill_color)

    return img


def generate_patterns(input: str, output: str) -> list[Pattern]:
    """
    Generate output patterns from input for ALL occurrences of output, starting from the right.

    Args:
        input: Input string.
        output: Output string.

    Returns:
        list[Pattern]: List of patterns from all occurrences, prioritizing rightmost matches.
    """
    patterns: list[Pattern] = []
    positions: list[int] = []

    # Find all occurrences of output in input (from right to left)
    start = len(input)
    while True:
        pos = input.rfind(output, 0, start)
        if pos == -1:
            break
        positions.append(pos)
        start = pos  # Continue searching to the left

    # Process positions (already in right-to-left order)
    for pos in positions:
        # Get the full before/after context for this occurrence
        before = input[:pos]
        after = input[pos + len(output) :]

        # Generate patterns of different delimiter lengths for this occurrence
        for delim_len in range(min(20, len(before), len(after)), 0, -1):
            pattern = Pattern(before=before[-delim_len:], after=after[:delim_len])
            if pattern not in patterns:  # Avoid duplicates
                patterns.append(pattern)

    return patterns


def parse_python_code(markdown: str) -> str:
    # Pattern to match code fences with optional language specification
    # Matches ```python, ```py, or just ``` followed by code and ending ```
    pattern = r"```(?:python|py)?\s*\n(.*?)\n```"

    matches = re.findall(pattern, markdown, re.DOTALL)
    if not matches:
        raise ValueError("No code found in markdown")

    return matches[0].strip()


class LimitOffsetTracker:
    def __init__(self, limit: int, offset: int):
        self.offset = offset
        self.limit = limit
        self.global_position = 0
        self.remaining_items = limit

    def slice(self, data: list) -> list:
        if not data:
            return []

        chunk_start = max(0, self.offset - self.global_position)
        chunk_end = min(len(data), chunk_start + self.remaining_items)
        self.global_position += len(data)

        if chunk_start < len(data):
            slice_data = data[chunk_start:chunk_end]
            if slice_data:
                self.remaining_items -= len(slice_data)
                return slice_data

        return []


def is_digit_value(value: Any) -> bool:
    """Check if value is a digit (for page, limit, offset)"""
    return isinstance(value, int) or (isinstance(value, str) and value.strip().isdigit())


def is_potential_cursor(value: str) -> bool:
    if re.match(r"^\d{4}-\d{2}-\d{2}[T ]\d{2}:\d{2}:\d{2}", value):  # ISO datetime format
        return True
    if len(value) < 8:
        return False
    return bool(re.match(r"^[A-Za-z0-9_\-+:.=/]+$", value))


def extract_potential_cursors(value: Any) -> list[str]:
    """Extract all potential cursors using regex patterns with proper boundaries"""
    value_str = str(value)
    extracted_values = []

    # Step 1: Check if whole string is a cursor
    if is_potential_cursor(value_str):
        return [value_str]

    # Step 2: Look for patterns
    patterns = [
        r'"([^"]+)"',  # Double quoted strings
        r"'([^']+)'",  # Single quoted strings
        r'\\"([^"]+)\\"',  # Double quoted strings
    ]

    for pattern in patterns:
        matches = re.findall(pattern, value_str)
        for match in matches:
            if is_potential_cursor(match):
                extracted_values.append(match)

    return list(set(extracted_values))


def get_potential_pagination_parameters(request: Request) -> dict[str, Any] | None:
    """
    Extract potential pagination parameters from request queries and post data.

    Args:
        request: Request object

    Returns:
        Dictionary of potential pagination parameters or None if none found
    """
    result = {}

    if request.queries:
        for k, v in request.queries.items():
            if is_digit_value(v) or bool(extract_potential_cursors(v)):
                result[k] = v

    if isinstance(request.post_data, dict):
        for k, v in request.post_data.items():
            if is_digit_value(v) or bool(extract_potential_cursors(v)):
                result[k] = v

    return result if result else None



from __future__ import annotations

import asyncio
import contextlib
import io
import os
from collections.abc import Callable
from json import dumps as json_dumps
from typing import Any, cast

from patchright.async_api import Browser
from pydantic import BaseModel

from strot import llm
from strot.analyzer import prompts
from strot.analyzer.schema import HTMLResponsePreprocessor, Point, Response, pagination_strategy
from strot.analyzer.schema.request import Request
from strot.analyzer.schema.source import Source
from strot.analyzer.utils import (
    draw_point_on_image,
    encode_image,
    extract_potential_cursors,
    generate_patterns,
    get_potential_pagination_parameters,
    parse_python_code,
    text_match_ratio,
)
from strot.browser import launch_browser
from strot.browser.tab import Tab
from strot.logging import LoggerType, get_logger, setup_logging
from strot.type_adapter import TypeAdapter

__all__ = ("analyze", "Analyzer")

setup_logging()

HEADERS_TO_IGNORE = {
    "accept-encoding",
    "host",
    "method",
    "path",
    "scheme",
    "version",
    "authority",
    "protocol",
    "content-length",
}


async def analyze(
    *,
    url: str,
    query: str,
    output_schema: type[BaseModel],
    max_steps: int = 30,
    browser: Browser | None = None,
    logger: LoggerType | None = None,
    page_load_timeout: float | None = None,
) -> Source | None:
    """
    Run analysis on the given URL using the given query.

    Args:
        url: The target web page URL to analyze.
        query: The query to use for analysis.
        output_schema: The output schema to extract data from the response.
        max_steps: The maximum number of steps to perform before exiting.
        browser: The browser to use.
        logger: The logger to use.
        page_load_timeout: The timeout for waiting for the page to load.
    """
    logger = logger or get_logger()

    async def run(browser: Browser):
        analyzer = Analyzer(logger=logger)
        browser_ctx = await browser.new_context(bypass_csp=True)
        tab = Tab(browser_ctx, load_timeout=page_load_timeout)
        try:
            logger.info("analysis", action="begin", status="pending", url=url, query=query)
            await tab.goto(url)
            source = await analyzer(tab, query, output_schema, max_steps)
            if not source:
                logger.info(
                    "analysis", action="end", status="failed", url=url, query=query, reason="No relevant request found"
                )
            else:
                logger.info(
                    "analysis",
                    action="end",
                    status="success",
                    url=url,
                    query=query,
                    source_url=source.request.url,
                )
        except Exception as e:
            logger.info("analysis", action="end", status="failed", url=url, query=query, reason=e)
            return None
        else:
            return source
        finally:
            with contextlib.suppress(Exception):
                await tab.reset()
                await browser_ctx.close()

    if browser is None:
        async with launch_browser("headed") as browser_instance:
            return await run(browser_instance)

    return await run(browser)


class Analyzer:
    def __init__(self, logger: LoggerType) -> None:
        self._logger = logger
        self._llm_client = llm.LLMClient(
            provider="anthropic",
            model="claude-sonnet-4-20250514",
            api_key=os.getenv("STROT_ANTHROPIC_API_KEY"),
            cost_per_1m_input=3.0,
            cost_per_1m_output=15.0,
        )

    async def request_llm_completion(
        self,
        event: str,
        input: llm.LLMInput,
        json: bool,
        validator: Callable[[str], Any],
    ) -> Any:
        self._logger.info(
            event,
            action="llm-completion",
            provider=self._llm_client.provider,
            model=self._llm_client.model,
            status="pending",
        )
        try:
            completion = await self._llm_client.get_completion(input, json=json)
            self._logger.info(
                event,
                action="llm-completion",
                provider=self._llm_client.provider,
                model=self._llm_client.model,
                status="success",
                result=completion.value,
                input_tokens=completion.input_tokens,
                output_tokens=completion.output_tokens,
                cost=self._llm_client.calculate_cost(completion.input_tokens, completion.output_tokens),
            )
            return validator(completion.value)
        except Exception as e:
            self._logger.info(
                event,
                action="llm-completion",
                provider=self._llm_client.provider,
                model=self._llm_client.model,
                status="failed",
                reason=e,
            )
            raise

    async def run_step(self, tab: Tab, query: str) -> Response | None:  # noqa: C901
        screenshot = await tab.plugin.take_screenshot(type="png")
        type_adapter = TypeAdapter(prompts.schema.StepResult)
        schema = type_adapter.generate_schema(drop_titles=True)
        schema.pop("description", None)
        prompt = prompts.ANALYZE_CURRENT_VIEW_PROMPT_TEMPLATE.render(
            output_schema=json_dumps(schema, indent=2), requirement=query
        )

        try:
            result = cast(
                prompts.schema.StepResult,
                await self.request_llm_completion(
                    event="run-step",
                    input=llm.LLMInput(prompt=prompt, image=screenshot),
                    json=True,
                    validator=lambda x: type_adapter.validate_json(x),
                ),
            )
        except Exception:
            return None

        def get_context(point: Point) -> bytes:
            image = draw_point_on_image(screenshot, point)
            buffer = io.BytesIO()
            image.save(buffer, format="PNG")
            return encode_image(buffer.getvalue())

        if sections := result.text_sections:
            response_to_return = None
            container = await tab.plugin.get_parent_container(sections)
            for response in tab.responses:
                score = text_match_ratio(sections, response.value)
                if score < 0.5:
                    continue

                if response.request.type == "ssr" and container:
                    response.preprocessor = HTMLResponsePreprocessor(element_selector=container)

                response_to_return = response
                break

            skip = False
            if container and (last_child := await tab.plugin.get_last_visible_child(container)):
                self._logger.info(
                    "run-step",
                    context=encode_image(screenshot),
                    step="skip-similar-content",
                    action="scroll",
                    target=last_child,
                )
                await tab.plugin.scroll_to_element(last_child)
                skip = True

            if point := result.close_overlay_popup_coords:
                self._logger.info(
                    "run-step",
                    context=get_context(point),
                    step="close-overlay-popup",
                    action="click",
                    point=point.model_dump(),
                )
                await tab.plugin.click_at_point(point)

            if skip or response_to_return:
                # If either similar element skip was performed or response was found, end the step
                return response_to_return

        if point := result.close_overlay_popup_coords:
            self._logger.info(
                "run-step",
                context=get_context(point),
                step="close-overlay-popup",
                action="click",
                point=point.model_dump(),
            )
            if await tab.plugin.click_at_point(point):
                return

        if point := result.load_more_content_coords:
            self._logger.info(
                "run-step",
                context=get_context(point),
                step="load-more-content",
                action="click",
                point=point.model_dump(),
            )
            if await tab.plugin.click_at_point(point):
                return

        if point := result.skip_to_content_coords:
            self._logger.info(
                "run-step", context=get_context(point), step="skip-to-content", action="click", point=point.model_dump()
            )
            if await tab.plugin.click_at_point(point):
                return

        self._logger.info(
            "run-step", context=encode_image(screenshot), step="fallback", action="scroll", target="next-view"
        )
        await tab.plugin.scroll_to_next_view()

    async def discover_relevant_response(self, tab: Tab, query: str, max_steps: int | MutableRange) -> Response | None:
        for step in range(0, max_steps) if isinstance(max_steps, int) else max_steps:
            try:
                self._logger.info("request-detection", step_count=step, action="run-step", status="pending")
                response = await self.run_step(tab, query)
            except Exception as e:
                self._logger.info("request-detection", step_count=step, action="run-step", status="failed", reason=e)
                response = None

            if not response:
                self._logger.info(
                    "request-detection",
                    step_count=step,
                    action="run-step",
                    status="failed",
                    reason="No response on this step.",
                )
                await asyncio.sleep(2.5)
                continue

            # Log detailed success information
            success_log_kwargs = {
                "request_type": response.request.type,
                "method": response.request.method,
                "url": response.request.url,
                "queries": response.request.queries,
                "data": response.request.post_data,
            }
            if response.preprocessor:
                success_log_kwargs["response_preprocessor"] = response.preprocessor.model_dump()

            self._logger.info(
                "request-detection", step_count=step, action="run-step", status="success", **success_log_kwargs
            )
            return response

    async def detect_pagination_keys_and_strategy(  # noqa: C901
        self, request: Request, *responses: Response
    ) -> tuple[prompts.schema.PaginationKeys, pagination_strategy.PaginationStrategy | None]:
        keys = prompts.schema.PaginationKeys()
        potential_pagination_parameters = get_potential_pagination_parameters(request)
        self._logger.info("pagination-detection", parameters=potential_pagination_parameters, status="pending")
        if not potential_pagination_parameters:
            self._logger.info(
                "pagination-detection",
                parameters=potential_pagination_parameters,
                status="failed",
                reason="No potential pagination parameters detected.",
            )
            return keys, None

        type_adapter = TypeAdapter(prompts.schema.PaginationKeys)
        schema = type_adapter.generate_schema(drop_titles=True)
        schema.pop("description", None)
        llm_input = llm.LLMInput(
            prompt=prompts.IDENTIFY_PAGINATION_KEYS_PROMPT_TEMPLATE.render(
                parameters=json_dumps(potential_pagination_parameters, indent=2),
                output_schema=json_dumps(schema, indent=2),
            )
        )

        for _ in range(3):
            try:
                keys = cast(
                    prompts.schema.PaginationKeys,
                    await self.request_llm_completion(
                        event="pagination-detection",
                        input=llm_input,
                        json=True,
                        validator=lambda x: type_adapter.validate_json(x),
                    ),
                )
                break
            except Exception:  # noqa: S112
                continue

        if not (keys.page_number_key or keys.cursor_key or keys.offset_key):
            self._logger.info(
                "pagination-detection",
                parameters=potential_pagination_parameters,
                status="failed",
                reason="No pagination keys detected.",
            )
            return keys, None

        page_param = None
        if keys.page_number_key:
            page_param = pagination_strategy.NumberParameter(
                key=keys.page_number_key,
                default_value=int(potential_pagination_parameters.get(keys.page_number_key, 0)),
            )

        cursor_param = None
        if (
            (cursor_key := keys.cursor_key)
            and (cursor := potential_pagination_parameters.get(cursor_key))
            and (potential_sub_cursors := extract_potential_cursors(cursor))
        ):
            best_match_count, best_response_text = 0, None

            for response in responses:
                response_parameters = get_potential_pagination_parameters(response.request)
                if response_parameters and set(response_parameters).issubset(potential_pagination_parameters):
                    match_count = sum(1 for v in potential_sub_cursors if v in response.value)
                    if match_count >= best_match_count:
                        best_match_count = match_count
                        best_response_text = response.value

            if best_response_text and best_match_count > 0:
                pattern_map = {}
                for value in potential_sub_cursors:
                    if patterns := generate_patterns(best_response_text, value):
                        pattern_map[value] = patterns

                if pattern_map:
                    cursor_param = pagination_strategy.CursorParameter(
                        key=cursor_key, default_value=str(cursor), pattern_map=pattern_map
                    )

        offset_param = None
        if keys.offset_key:
            offset_param = pagination_strategy.NumberParameter(
                key=keys.offset_key,
                default_value=int(potential_pagination_parameters.get(keys.offset_key, 0)),
            )

        limit_param = None
        if keys.limit_key:
            limit_param = pagination_strategy.NumberParameter(
                key=keys.limit_key,
                default_value=int(potential_pagination_parameters.get(keys.limit_key, 1)),
            )

        if page_param or offset_param or cursor_param:
            strategy = pagination_strategy.PaginationStrategy(
                page=page_param, limit=limit_param, offset=offset_param, cursor=cursor_param
            )
            self._logger.info(
                "pagination-detection",
                parameters=potential_pagination_parameters,
                status="success",
                strategy=strategy.model_dump(exclude_none=True),
            )
            return keys, strategy

        self._logger.info(
            "pagination-detection",
            parameters=potential_pagination_parameters,
            status="failed",
            reason="Could not determine the strategy.",
        )
        return keys, None

    async def get_extraction_code_and_default_limit(
        self, response: Response, output_schema: type[BaseModel]
    ) -> tuple[str | None, int | None]:
        self._logger.info(
            "code-generation",
            response_length=len(response.value),
            preprocessor=response.preprocessor.model_dump() if response.preprocessor else None,
            status="pending",
        )
        response_text = response.value
        if response.preprocessor:
            response_text = response.preprocessor.run(response_text) or response_text

        # Every schema will be treated as If we're extracting a list of items
        type_adapter = TypeAdapter(list[output_schema])
        schema = type_adapter.generate_schema(drop_titles=True)
        schema.pop("description", None)
        llm_input = llm.LLMInput(
            prompt=prompts.GENERATE_EXTRACTION_CODE_PROMPT_TEMPLATE.render(
                output_schema=json_dumps(schema, indent=2),
                api_response=response_text,
            )
        )

        def validate(value: str) -> tuple[str, int]:
            if code := parse_python_code(value):
                namespace = {}
                exec(code, namespace)  # noqa: S102
                data = type_adapter.validate_python(namespace["extract_data"](response_text))
                return code, len(data)

            raise ValueError("Code parsing failed")

        for _ in range(3):
            try:
                code, default_limit = cast(
                    tuple[str, int],
                    await self.request_llm_completion(
                        event="code-generation",
                        input=llm_input,
                        json=False,
                        validator=validate,
                    ),
                )
                self._logger.info(
                    "code-generation",
                    response_length=len(response.value),
                    preprocessor=response.preprocessor.model_dump() if response.preprocessor else None,
                    status="success",
                    code=code,
                    default_limit=default_limit,
                )
                return code, default_limit  # noqa: TRY300
            except Exception:  # noqa: S112
                continue

        self._logger.info(
            "code-generation",
            response_length=len(response.value),
            preprocessor=response.preprocessor.model_dump() if response.preprocessor else None,
            status="failed",
            reason="Could not determine default limit",
        )
        return None, None

    async def __call__(
        self,
        tab: Tab,
        query: str,
        output_schema: type[BaseModel],
        max_steps: int = 30,
    ) -> Source | None:
        steps = MutableRange(0, max_steps)
        response, strategy = None, None
        while True:
            self._logger.info("analysis", action="request-detection", status="pending")
            response = await self.discover_relevant_response(tab, query, steps)
            if not response:
                self._logger.info(
                    "analysis", action="request-detection", status="failed", reason="No relevant response detected."
                )
                break

            self._logger.info("analysis", action="request-detection", status="success")

            self._logger.info("analysis", action="pagination-detection", status="pending")
            _, strategy = await self.detect_pagination_keys_and_strategy(response.request, *tab.responses)
            if strategy is None:
                self._logger.info(
                    "analysis", action="pagination-detection", status="failed", reason="No pagination detected."
                )
                await tab.plugin.scroll_to_next_view()
                continue

            self._logger.info("analysis", action="pagination-detection", status="success")
            break

        if not response:
            return None

        self._logger.info("analysis", action="code-generation", status="pending")
        code, default_limit = await self.get_extraction_code_and_default_limit(response, output_schema)
        if not code:
            self._logger.info(
                "analysis", action="code-generation", status="failed", reason="LLM failed to generate code."
            )
        else:
            self._logger.info("analysis", action="code-generation", status="success")

        for key in list(response.request.headers):
            if key.lstrip(":").lower() in HEADERS_TO_IGNORE:
                response.request.headers.pop(key)

        return Source(
            request=response.request,
            pagination_strategy=strategy,
            response_preprocessor=response.preprocessor,
            extraction_code=code,
            default_limit=default_limit or 1,
        )


class MutableRange:
    def __init__(self, start: int, stop: int, step: int = 1):
        if step == 0:
            raise ValueError("step argument must not be zero")
        self._current = start
        self._stop = stop
        self._step = step

    def __iter__(self):
        if self._step > 0:
            # Positive step: current < stop
            while self._current < self._stop:
                value = self._current
                self._current += self._step
                yield value
        else:
            # Negative step: current > stop
            while self._current > self._stop:
                value = self._current
                self._current += self._step  # Adding negative step decreases current
                yield value



import contextlib
from urllib.parse import parse_qsl, urlparse

from patchright.async_api import BrowserContext, Page
from patchright.async_api import Response as InterceptedResponse

from strot.analyzer.schema import Response
from strot.analyzer.schema.request import Request
from strot.browser.plugin import Plugin

EXCLUDE_KEYWORDS = {"analytics", "telemetry", "events", "collector", "track", "collect"}


class Tab:
    def __init__(
        self,
        browser_context: BrowserContext,
        viewport_size: dict[str, int] | None = None,
        load_timeout: float | None = None,
    ) -> None:
        self._browser_context = browser_context
        self._viewport_size = viewport_size or {"width": 1280, "height": 800}
        self._load_timeout = load_timeout

        self._page: Page | None = None
        self._page_headers: dict[str, str] | None = None
        self._responses: list[Response] = []
        self._plugin: Plugin | None = None

    def is_empty(self) -> bool:
        return self._page is None

    async def goto(self, url: str) -> None:
        await self.reset()

        self._page = await self._browser_context.new_page()
        self._page.on("response", self._handle_ajax_response)

        await self._page.set_viewport_size(self._viewport_size)

        with contextlib.suppress(Exception):
            response = await self._page.goto(url, timeout=self._load_timeout, wait_until="domcontentloaded")
            if response:
                self._page_headers = await response.request.all_headers()

        await self._page.wait_for_timeout(5000)
        self._page.on("load", self._handle_server_side_rendering)

        self._plugin = Plugin(self._page)

    @property
    def responses(self) -> list[Response]:
        return self._responses

    @property
    def plugin(self) -> Plugin | None:
        return self._plugin

    async def reset(self) -> None:
        if self._page:
            with contextlib.suppress(Exception):
                await self._page.close()
            self._page = None
            self._page_headers = None
            self._plugin = None
            self._responses.clear()

    async def _handle_ajax_response(self, response: InterceptedResponse) -> None:
        rsc_type = response.request.resource_type
        if rsc_type not in ("xhr", "fetch"):
            return

        url = urlparse(response.request.url)
        if url.scheme.lower() not in ("http", "https"):
            return

        clean_url = (url.netloc + url.path).lower()
        if any(w in clean_url for w in EXCLUDE_KEYWORDS):
            return

        with contextlib.suppress(Exception):
            self._responses.append(
                Response(
                    value=await response.text(),
                    request=Request(
                        method=response.request.method,
                        url=f"{url.scheme}://{url.netloc}{url.path}",
                        queries=dict(parse_qsl(url.query)),
                        type="ajax",
                        headers=await response.request.all_headers(),
                        post_data=response.request.post_data_json,
                    ),
                )
            )

    async def _handle_server_side_rendering(self, page: Page) -> None:
        parsed_url = urlparse(page.url)
        if parsed_url.scheme.lower() not in ("http", "https"):
            return

        self._responses.append(
            Response(
                value=await page.content(),
                request=Request(
                    method="GET",
                    url=f"{parsed_url.scheme}://{parsed_url.netloc}{parsed_url.path}",
                    type="ssr",
                    queries=dict(parse_qsl(parsed_url.query)),
                    headers=self._page_headers or {},
                    post_data=None,
                ),
            )
        )



from collections.abc import Callable
from typing import TypedDict

from pydantic import BaseModel, model_validator

from strot.analyzer.schema import Pattern, ResponsePreprocessorT
from strot.analyzer.schema.request import Request, RequestException
from strot.analyzer.utils import LimitOffsetTracker

__all__ = ("PaginationStrategy", "NumberParameter", "CursorParameter")


class NumberParameter(TypedDict):
    key: str
    default_value: int


class CursorParameter(TypedDict):
    key: str
    default_value: str
    pattern_map: dict[str, list[Pattern]]


class PaginationStrategy(BaseModel):
    page: NumberParameter | None = None
    cursor: CursorParameter | None = None
    limit: NumberParameter | None = None
    offset: NumberParameter | None = None

    @model_validator(mode="after")
    def validate_required_fields(self):
        if not self.page and not self.offset and not self.cursor:
            raise ValueError("Either page, offset or cursor parameter is required")
        return self

    async def generate_data(
        self,
        *,
        request: Request,
        tracker: LimitOffsetTracker,
        response_preprocessor: ResponsePreprocessorT | None = None,
        extract_fn: Callable[[str], list],
        default_limit: int = 1,
    ):
        gen_fn = self._get_gen_fn()
        async for data in gen_fn(request, tracker, response_preprocessor, extract_fn, default_limit):
            yield data

    def _get_gen_fn(self):
        """Return the pagination combination in priority order"""
        if self.limit and self.offset:
            return self._generate_limit_offset_data
        elif self.page and self.limit:
            return self._generate_page_limit_data
        elif self.page and self.offset:
            return self._generate_page_offset_data
        elif self.cursor:
            return self._generate_cursor_data
        else:
            return self._generate_page_limit_data

    @staticmethod
    async def _make_request(
        request: Request,
        state: dict,
        response_preprocessor: ResponsePreprocessorT | None,
    ) -> tuple[str, str | None]:
        response = await request.make(state=state)
        original_text = await response.text()
        if response_preprocessor:
            return original_text, response_preprocessor.run(original_text)
        return original_text, original_text

    async def _generate_limit_offset_data(
        self,
        request: Request,
        tracker: LimitOffsetTracker,
        response_preprocessor: ResponsePreprocessorT | None,
        extract_fn: Callable[[str], list],
        default_limit: int,
    ):
        """Generate data using limit/offset pagination"""
        page_size = tracker.limit if self.limit else default_limit

        first_request = True
        last_response = None
        used_fallback = False  # Track if we've already used the default limit fallback

        # Initialize tracker's global position to start at the beginning of our first request
        tracker.global_position = tracker.offset

        while tracker.remaining_items > 0:
            state = {
                self.offset["key"]: str(tracker.global_position),
            }
            if self.limit:
                state[self.limit["key"]] = str(page_size)

            try:
                _, response = await self._make_request(request, state, response_preprocessor)
            except RequestException as e:
                # On first 400 error, try default limit if limit key is available and we haven't used fallback yet
                if e.status_code == 400 and self.limit and not used_fallback:
                    page_size = default_limit
                    state[self.limit["key"]] = str(page_size)
                    used_fallback = True
                    _, response = await self._make_request(request, state, response_preprocessor)
                else:
                    raise

            if not response or response == last_response:
                break
            last_response = response
            data = extract_fn(response)

            # Detect API's actual limit on first request
            if first_request:
                if len(data) == 0:
                    # If first request returns no data, API doesn't support this limit
                    break
                elif self.limit and len(data) < page_size:
                    page_size = len(data)
                first_request = False

            if slice_data := tracker.slice(data):
                yield slice_data
            else:
                break

    async def _generate_page_limit_data(  # noqa: C901
        self,
        request: Request,
        tracker: LimitOffsetTracker,
        response_preprocessor: ResponsePreprocessorT | None,
        extract_fn: Callable[[str], list],
        default_limit: int,
    ):
        """Generate data using page/limit pagination"""
        page_base = await self._detect_page_base(self.page, request, response_preprocessor, extract_fn)

        # Determine the page size (items per page)
        page_size = default_limit
        if self.limit:
            # Try to use user's requested limit as page size, fall back to default
            try:
                test_state = {self.page["key"]: str(page_base), self.limit["key"]: str(tracker.limit)}
                _, response = await self._make_request(request, test_state, response_preprocessor)
                if response and (test_data := extract_fn(response)):
                    page_size = min(len(test_data), tracker.limit)
            except RequestException:
                page_size = default_limit

        # Calculate which pages we need to fetch
        start_page = page_base + (tracker.offset // page_size)

        current_page = start_page
        last_response = None
        used_fallback = False  # Track if we've already used the default limit fallback

        tracker.global_position = tracker.offset
        while tracker.remaining_items > 0:
            # Prepare request state
            state = {self.page["key"]: str(current_page)}
            if self.limit:
                state[self.limit["key"]] = str(page_size)

            try:
                _, response = await self._make_request(request, state, response_preprocessor)
            except RequestException as e:
                # On first 400 error, try default limit if limit key is available and we haven't used fallback yet
                if e.status_code == 400 and self.limit and not used_fallback:
                    page_size = default_limit
                    state[self.limit["key"]] = str(page_size)
                    used_fallback = True
                    _, response = await self._make_request(request, state, response_preprocessor)
                else:
                    break

            # Check for identical responses (end of data)
            if not response or response == last_response:
                if current_page == 1:  # both page 0 and 1 can return the same response
                    current_page += 1
                    continue
                break

            last_response = response
            data = extract_fn(response)
            if not data:
                break

            # Use tracker.slice to handle offset/limit logic
            if slice_data := tracker.slice(data):
                yield slice_data
            else:
                break

            current_page += 1

    async def _generate_page_offset_data(
        self,
        request: Request,
        tracker: LimitOffsetTracker,
        response_preprocessor: ResponsePreprocessorT | None,
        extract_fn: Callable[[str], list],
        default_limit: int,
    ):
        """Generate data using page/offset pagination"""
        page_base = await self._detect_page_base(self.page, request, response_preprocessor, extract_fn)

        # This is tricky - we need to figure out how page and offset interact
        # Common patterns:
        # 1. page=N, offset=M means "start at page N, then skip M items within that page"
        # 2. page=N, offset=M means "page N of results, starting from global offset M"

        # We'll assume pattern 1 (offset within page) for now
        # In a real implementation, you'd need to test the API behavior

        # Calculate starting page based on user offset and estimated page size
        estimated_page_size = default_limit
        start_page = page_base + (tracker.offset // estimated_page_size)
        end_item = tracker.offset + tracker.limit
        end_page = page_base + ((end_item - 1) // estimated_page_size)
        offset_within_page = tracker.offset % estimated_page_size

        current_page = start_page
        last_response = None

        # Set initial global position for tracker based on the starting page
        tracker.global_position = (start_page - page_base) * estimated_page_size

        while tracker.remaining_items > 0 and current_page <= end_page:
            state = {
                self.page["key"]: str(current_page),
                self.offset["key"]: str(self.offset["default_value"] + offset_within_page),
            }

            try:
                _, response = await self._make_request(request, state, response_preprocessor)
            except RequestException:
                break

            if not response or response == last_response:
                if current_page == 1:  # both page 0 and 1 can return the same response
                    current_page += 1
                    continue
                break
            last_response = response

            data = extract_fn(response)
            if not data:
                break

            # Use tracker.slice to handle offset/limit logic
            if slice_data := tracker.slice(data):
                yield slice_data
            else:
                break

            # For subsequent pages, no offset within page
            if current_page > start_page:
                offset_within_page = 0

            current_page += 1

    async def _generate_cursor_data(  # noqa: C901
        self,
        request: Request,
        tracker: LimitOffsetTracker,
        response_preprocessor: ResponsePreprocessorT | None,
        extract_fn: Callable[[str], list],
        default_limit: int = 1,
    ):
        all_cursors = []
        start_cursor = await get_start_cursor(request, self.cursor)
        if start_cursor:
            all_cursors.append(start_cursor)
        state = {self.cursor["key"]: start_cursor}
        if self.page:
            state[self.page["key"]] = str(
                await self._detect_page_base(self.page, request, response_preprocessor, extract_fn)
            )
        if self.limit:
            state[self.limit["key"]] = str(tracker.limit)

        first_request = True
        last_response = None
        if self.offset:
            tracker.global_position = tracker.offset
        while tracker.remaining_items > 0:
            if self.offset:
                state[self.offset["key"]] = str(tracker.global_position)
            try:
                orig, response = await self._make_request(request, state, response_preprocessor)
            except RequestException as e:
                if e.status_code == 400 and self.limit and first_request:
                    state[self.limit["key"]] = str(default_limit)
                    orig, response = await self._make_request(request, state, response_preprocessor)
                else:
                    raise
            if not response or response == last_response:
                if self.page:
                    current_page = int(state[self.page["key"]])
                    if current_page == 1:  # both page 0 and 1 can return the same response
                        state[self.page["key"]] = str(current_page + 1)
                        continue
                break
            last_response = response
            data = extract_fn(response)

            # Detect API's actual limit on first request
            if self.limit and first_request and len(data) < tracker.limit:
                state[self.limit["key"]] = str(len(data))
                first_request = False

            if slice_data := tracker.slice(data):
                yield slice_data

            next_cursor = extract_cursor(orig, self.cursor)
            if next_cursor is None or next_cursor in all_cursors:
                break
            all_cursors.append(next_cursor)
            state[self.cursor["key"]] = next_cursor
            if self.page:
                state[self.page["key"]] = str(int(state[self.page["key"]]) + 1)

    async def _detect_page_base(
        self,
        page_parameter: NumberParameter,
        request: Request,
        response_preprocessor: ResponsePreprocessorT | None,
        extract_fn: Callable[[str], list],
    ) -> int:
        """Detect if API uses 0-based or 1-based page numbering by testing page 0"""
        try:
            _, response = await self._make_request(request, {page_parameter["key"]: "0"}, response_preprocessor)
            if response and extract_fn(response):
                return 0  # Page 0 works, API is 0-based
        except RequestException:
            pass
        return 1  # Page 0 failed, API is 1-based


async def get_start_cursor(request: Request, cursor_parameter: CursorParameter) -> str | None:
    """Determine the starting cursor value for pagination."""

    # Try 1: No cursor (some APIs support starting without cursor)
    try:
        _ = await (await request.make(state={cursor_parameter["key"]: None})).text()
    except RequestException:
        pass
    else:
        return None

    # Try 2: Replace sub cursors with null to create a "first page" cursor
    try:
        cursor = cursor_parameter["default_value"]
        for value in cursor_parameter["pattern_map"]:
            # Replace the cursor value with "null", handling quoted values properly
            pos = cursor.find(value)
            if pos != -1:
                # Check for quotes around the value
                if pos > 0 and cursor[pos - 1] == '"':
                    value = f'"{value}"'
                    i = 2
                    while pos >= i and cursor[pos - i] == "\\":
                        value = f"\\{value}\\"
                        i += 1

                cursor = cursor.replace(value, "null")

        _ = await (await request.make(state={cursor_parameter["key"]: cursor})).text()
    except RequestException:
        pass
    else:
        return cursor

    # Fallback: Use default cursor as starting cursor
    return cursor_parameter["default_value"]


def extract_cursor(response_text: str, cursor_parameter: CursorParameter) -> str | None:
    """Extract cursor from response using pattern map"""
    cursor_values = {}

    for value, patterns in cursor_parameter["pattern_map"].items():
        if not patterns:
            # Constant value that doesn't need updating
            cursor_values[value] = value
            continue

        for pattern in patterns:
            if (output := pattern.test(response_text)) and len(output) == len(value):
                cursor_values[value] = output
                break
        else:
            # If we can't find a pattern for this value, cursor extraction failed
            return None

    if not cursor_values:
        return None

    # Reconstruct cursor with new values
    new_cursor = cursor_parameter["default_value"]
    for old_value in sorted(cursor_values.keys(), key=len, reverse=True):
        new_cursor = new_cursor.replace(old_value, cursor_values[old_value])

    return new_cursor



import base64
import io
import json
import logging
from enum import Enum
from inspect import currentframe
from pathlib import Path
from typing import Any, Protocol

import structlog
from PIL import Image
from structlog.processors import CallsiteParameter

__all__ = (
    "setup_logging",
    "get_logger",
    "LogLevel",
    "LoggerType",
    "BaseHandlerConfig",
)

LoggerType = structlog.stdlib.BoundLogger


class LogLevel(Enum):
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL


class BaseHandlerConfig(Protocol):
    """Protocol for handler configurations."""

    def get_handler(self, logger_name: str) -> logging.Handler:
        """Get a configured logging handler."""
        ...


_level = LogLevel.INFO


def setup_logging(
    level: LogLevel | None = None,
    overrides: dict[str, LogLevel | None] | None = None,
) -> None:
    """
    Initialize the logger.

    Args:
        level: Logging level. Defaults to INFO.
        overrides: Logger names mapped to their logging levels. If level value is None, the logger will be disabled.
    """
    if level is not None:
        global _level
        _level = level

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(ConsoleFormatter())

    logging.basicConfig(
        level=_level.value,
        format="%(message)s",
        handlers=[console_handler],
    )

    structlog.configure(
        processors=[
            structlog.contextvars.merge_contextvars,
            structlog.processors.format_exc_info,
            structlog.stdlib.add_log_level,
            structlog.processors.TimeStamper(fmt="iso"),
            structlog.processors.CallsiteParameterAdder(parameters=[CallsiteParameter.FUNC_NAME]),
            structlog.stdlib.filter_by_level,
            structlog.processors.JSONRenderer(),
        ],
        context_class=dict,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )

    for logger_name, log_level in (overrides or {}).items():
        lg = logging.getLogger(logger_name)
        if log_level is None:
            lg.disabled = True
        else:
            lg.setLevel(log_level.value)


def get_logger(name: str | None = None, /, *handler_config: BaseHandlerConfig, **initial_values: Any) -> LoggerType:
    """
    Get a logger instance.

    Args:
        name: Logger name. Defaults to the name of the file where it is called.
        handler_config: Logging handler configurations.
        initial_values: Initial values to add to the logger context.

    Returns:
        LoggerType: A logger instance.
    """
    if name is None:
        try:
            frame = currentframe()
            source = frame.f_back.f_code.co_filename
            name = Path(source).stem
            if name == "__init__":
                name = Path(source).parent.stem
        except Exception:
            name = "unknown"

    logger: LoggerType = structlog.get_logger(name, **initial_values)

    for config in handler_config:
        logger.addHandler(config.get_handler(name))

    return logger


class ConsoleFormatter(logging.Formatter):
    def format(self, record):  # noqa: C901
        try:
            data = json.loads(record.getMessage())
        except json.JSONDecodeError:
            data = {"message": record.getMessage()}

        message_parts = [data.pop("level", record.levelname).upper(), f"event={data.pop("event", "unknown")!r}"]
        for k, v in data.items():
            if v is None:
                continue
            if k in ("func_name", "msg", "message", "timestamp"):
                message_parts.append(v)
            elif isinstance(v, Exception):
                message_parts.append(f"{v.__class__.__name__}: {v!s}")
            elif isinstance(v, float):
                message_parts.append(f"{k}={v:.2f}")
            elif isinstance(v, str):
                if "\n" in v:
                    message_parts.append(f"{k}='''\n{v.strip("\n")}\n'''")
                    continue
                try:
                    if Image.open(io.BytesIO(base64.b64decode(v))).format:
                        message_parts.append(f"{k}={v[:100]}...")
                        continue
                except Exception:  # noqa: S110
                    pass
                message_parts.append(f"{k}={v!r}")
            else:
                message_parts.append(f"{k}={v!r}")

        if record.exc_info:
            message_parts.append(self.formatException(record.exc_info))

        return " | ".join(message_parts)



from pathlib import Path

from jinja2 import Environment, FileSystemLoader

from . import schema

__all__ = (
    "ANALYZE_CURRENT_VIEW_PROMPT_TEMPLATE",
    "GENERATE_EXTRACTION_CODE_PROMPT_TEMPLATE",
    "IDENTIFY_PAGINATION_KEYS_PROMPT_TEMPLATE",
    "schema",
)

_env = Environment(loader=FileSystemLoader(Path(__file__).parent / "templates"))  # noqa: S701

ANALYZE_CURRENT_VIEW_PROMPT_TEMPLATE = _env.get_template("analyze_current_view.jinja")

IDENTIFY_PAGINATION_KEYS_PROMPT_TEMPLATE = _env.get_template("identify_pagination_keys.jinja")

GENERATE_EXTRACTION_CODE_PROMPT_TEMPLATE = _env.get_template("generate_extraction_code.jinja")



import asyncio
import inspect
from contextlib import AbstractAsyncContextManager, asynccontextmanager, suppress
from typing import Any, Literal, overload

from patchright.async_api import Browser, async_playwright

__all__ = ("launch_browser",)


@overload
def launch_browser(mode: Literal["headed", "headless"], /) -> "AbstractAsyncContextManager[ResilientBrowser]":
    """
    Launch a browser instance.

    Args:
        mode: browser mode to use. Can be "headed" or "headless".

    Yields:
        A browser instance.
    """


@overload
def launch_browser(ws_url: str, /) -> "AbstractAsyncContextManager[ResilientBrowser]":
    """
    Launch a browser instance.

    Args:
        ws_url: The WebSocket URL to the browser instance.

    Yields:
        A browser instance.
    """


@asynccontextmanager
async def launch_browser(mode_or_ws_url: str, /):
    """
    Launch a browser instance.

    Args:
        mode_or_ws_url: The browser mode or WebSocket URL.

    Yields:
        A browser instance.
    """
    browser = ResilientBrowser(mode_or_ws_url)
    try:
        await browser.connect()
        yield browser
    finally:
        with suppress(Exception):
            await browser.close()


class ResilientBrowser:
    """Browser wrapper that automatically reinitializes the browser instance."""

    def __init__(self, mode_or_ws_url: str, max_retries: int = 3):
        self.mode_or_ws_url = mode_or_ws_url
        self.max_retries = max_retries
        self._browser: Browser | None = None
        self._playwright = None
        self._lock = asyncio.Lock()
        self._initializing = False

    async def _get_instance(self) -> Browser:
        """Get browser instance with connection check."""
        async with self._lock:
            # Check if current browser is still connected
            if await self.is_connected():
                return self._browser

            # Prevent multiple initialization attempts
            if self._initializing:
                # Wait for ongoing initialization
                while self._initializing:
                    await asyncio.sleep(0.1)
                if await self.is_connected():
                    return self._browser

            self._initializing = True
            try:
                # Cleanup old instances
                await self._close_browser()

                # Create new playwright instance if needed
                if not self._playwright:
                    self._playwright = await async_playwright().__aenter__()

                # Launch browser based on mode
                if self.mode_or_ws_url == "headed":
                    browser = await self._playwright.chromium.launch(headless=False)
                elif self.mode_or_ws_url == "headless":
                    browser = await self._playwright.chromium.launch(headless=True)
                else:
                    # Try CDP connection first, fallback to WebSocket
                    try:
                        browser = await self._playwright.chromium.connect_over_cdp(self.mode_or_ws_url)
                    except Exception:
                        browser = await self._playwright.chromium.connect(self.mode_or_ws_url)

                self._browser = browser

                return browser
            finally:
                self._initializing = False

    async def is_connected(self) -> bool:
        """Check if browser is connected."""
        try:
            if not self._browser:
                return False
            return await self._browser.is_connected()
        except Exception:
            return False

    async def connect(self) -> None:
        """Connect to the browser."""
        await self._get_instance()

    def _is_method(self, name: str) -> bool:
        """Check if an attribute is a method using inspect."""
        try:
            # Get the attribute from the class
            attr = getattr(Browser, name, None)
            if attr is None:
                return True  # Assume method if not found

            # Check if it's a method or function
            return (
                inspect.ismethod(attr)
                or inspect.isfunction(attr)
                or inspect.iscoroutinefunction(attr)
                or callable(attr)
            )
        except Exception:
            return True  # Assume method on error

    def __getattr__(self, name: str) -> Any:
        """Proxy attribute access with resilient logic."""

        # Check if it's a method or property
        if self._is_method(name):
            # Return async wrapper for methods
            async def method_wrapper(*args, **kwargs):
                for attempt in range(self.max_retries + 1):
                    try:
                        browser = await self._get_instance()
                        attr = getattr(browser, name)

                        if inspect.iscoroutinefunction(attr):
                            return await attr(*args, **kwargs)
                        else:
                            return attr(*args, **kwargs)

                    except Exception:
                        if attempt < self.max_retries:
                            # Force reconnection
                            async with self._lock:
                                await self._close_browser()
                            await asyncio.sleep(0.5 * (attempt + 1))
                        else:
                            raise

            return method_wrapper

        # For properties, try to return directly if connected
        if self._browser is None:
            raise AttributeError("Browser not connected. Call '.connect()' first.")

        return getattr(self._browser, name)

    async def new_page(self, **kwargs):
        """Create a new page with retry logic."""
        browser = await self._get_instance()
        return await browser.new_page(**kwargs)

    async def new_context(self, **kwargs):
        """Create a new context with retry logic."""
        browser = await self._get_instance()
        return await browser.new_context(**kwargs)

    async def _close_browser(self):
        """Close the browser."""
        if self._browser:
            with suppress(Exception):
                await self._browser.close()
            self._browser = None

    async def close(self):
        """Close the browser and cleanup."""
        async with self._lock:
            await self._close_browser()

            if self._playwright:
                with suppress(Exception):
                    await self._playwright.__aexit__(None, None, None)
                self._playwright = None
